! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: binadapt_mod
!
! Authors:  Brendan Perry - NASA/GSFC (brendan.h.perry@nasa.gov)
!            Comprehensive rewrite, merger and modularization for SAS-21
!           Steven L. Snowden - NASA/GSFC (original algorithm)
!           Kip Kuntz - NASA/GSFC/JHU (original algorithm)
!
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: binadapt_mod.f90,v 1.2 2009/02/20 13:27:08 aibarra Exp $
!-
!
!  Makes adaptively smoothed and/or binned images
!
!  BHP: Part of the Extended Source Analysis Software collection of SAS tasks.
!  This task is now a merged binimage and adapt from pre-SAS-21 ESAS. The two
!  tasks shared 90% of the same functionality, so they were merged and the differences
!  are handled by the parameters. Adapt created adaptively smoothed, exposure corrected,
!  and  background subtracted images. 
!
!  INPUT:
!    prefix inst+expid OR comb to use combimage inputs
!     OLD: mos1S001-obj-im-300-1100.fits from mosspectra
!     NEW: mos1S001-fovimsky-300-1100.fits from mosspectra
!    elow: Lower energy limit for the energy band in eV.
!    ehigh: Upper energy limit for the energy band in eV.
!    withqdp: subtract model particle (QPB) background (from mosback)?
!    withprot: subtract soft proton background (from task proton)?
!    withprot: subtract swcx background (from task swcx)?
!    withmask: apply additional masking (from task cheese)?
!    maskfile: name of mask to apply (defaults depending on infile).
!    withbinning: apply binning?
!    binfactor: 1, 2, 4 should this be hardcoded or attempt ANY binfactor???!!!
!    withsmoothing: apply smoothing?
!    smoothfactor: Number of counts to accumulate for smoothing.
!    maskthresh: Scale factor for excluding regions based on a mask image.
! Files used:
!    QDP map
!    Proton sky map (e.g. mos1S001-swcximsky)
!
!  OUTPUT:
!    Adaptively smoothed image and optional radial profile QDP plot
!    filenames determined by infile input unless outfile is specified.
!
module binadapt_mod

  use dal
  use param 
  use errorHandling
  use esas_mod      ! in goflib, formerly in subsn (SS)
  use gofutils_mod  ! in goflib, formerly in marq_mod
  use gofcoords_mod ! in goflib, formerly in subsn (SS)
  use gofio_mod     ! in goflib, formerly in subsn (SS)

  implicit none 

  private

  character(len=*), parameter :: &
       prefixpar    = "prefix", &         ! inst + expid (1S001/2S002/S003/comb)
       elowpar      = "elow", &           ! lower energy
       ehighpar     = "ehigh", &          ! upper energy 
       partpar      = "withpartbkg", &    ! QPB particle background
       sppar        = "withspbkg", &      ! soft proton background
       swcxpar      = "withswcxbkg", &    ! swcx background
       maskpar      = "withmask", &       ! perform masking?
       maskfilepar  = "maskfile",&        ! mask filename 
       withbinpar   = "withbinning", &    ! perform binning? 
       binfactpar   = "binfactor", &      ! binfactor
       withsmoopar  = "withsmoothing", &  ! perform smoothing
       smoocntspar  = "smoothcounts", &   ! smoothingcounts
       tmpar        = "maskthresh" !, &     ! thresholdmasking
  !       plotspar     = "withplots"         ! create QDP radial and size plots?

  public binadapt

contains

  !---------------------------------------------------------------------
  subroutine binadapt
    !---------------------------------------------------------------------

    type(DataSetT) :: inIMG_S, outADAPT_S, outRATE_S, outSIZE_S, outSIGMA_S, &
         inMASK_S, inQPB_S, inPROT_S, inSWCX_S, pnOOT_S, inEXP_S, inBKG_S
    type(ArrayT)   :: inIMG_A, outADAPT_A, outRATE_A, outSIZE_A, outSIGMA_A, &
         inMASK_A, inQPB_A, inPROT_A, inSWCX_A, pnOOT_A, inEXP_A, inBKG_A
    type(TableT)   :: pnOOT_T, inPROT_T, inSWCX_T, inBKG_T, &
         inEXP_T, inIMG_T, inMASK_T
    character (len=256) :: inIMG_F, outADAPT_F, outRATE_F, outSIZE_F, outSIGMA_F, &
         inMASK_F, inQPB_F, inPROT_F, inSWCX_F, pnOOT_F, inEXP_F, inBKG_F, &
         outSIZEQ_F, outRADQ_F
    character (len=25) :: smode
    character (len=10) :: obsid,  cunitO(2), ctypeO(2), radecsysO, cunit(2), &
         ctype(2), cunitE(2), ctypeE(2), radecsysE, radecsys, style
    character (len=7)  :: name
    character (len=6)  :: filter
    character (len=5)  :: prefix, expid, celow, cehigh, exptype, suffix
    character (len=2)  :: inst
    integer :: xSize=900, ySize=900, xSizeD, ySizeD
    integer :: i, j, k, ii, jj, i1, i2, j1, j2, jj1, jj2
    integer :: status, ios, ibin, ir, irr, ioff, outlun, xdim, ydim, dims(2)
    integer :: type, frmt, revl, nR, el, eh, elow, ehigh
    integer :: binfact, typeA
    integer, allocatable, dimension (:) :: n
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: aDI32, inIMG_JP, pnOOT_JP,&
         inMASK_JP
    !
    logical :: doit, oot, warndie, withpart, withprot, withswcx, withmask, withbin, &
         withsmoo, withoot, withcomb
    !
    real :: backO, areaO, expoO, backD, areaD, expoD, smoofact
    real :: imcounts, qpbcounts, expcounts, protcounts, swcxcounts
    real :: maskthresh
    real :: xcen, ycen ! why reals???!!!
    real :: b, blast, c, clast, e, elast, ex, num, numlast, ootsca, r, &
         rad, s, sca, sn, temp, t, tlim, trad, xmax, ymax
    real(kind=double) :: crvalO(2), cdeltO(2), crpixO(2), crotaO(2), &
         ra_pntO, dec_pntO, pa_pntO, crval(2), cdelt(2), crpix(2), crota(2), &
         ra_pnt, dec_pnt, pa_pnt, crvalE(2), cdeltE(2), crpixE(2), crotaE(2), &
         ra_pntE, dec_pntE, pa_pntE
    real, dimension(:), allocatable :: bck, cts, expos, x, y, y2, ys ! why 50???!!!
    real, dimension(:), allocatable :: xh, yh ! why 10000/1000 in orig code???!!!
    real, dimension(:,:), allocatable :: mask, expose, back, counts, ootcounts, &
         prot, swcx, pnOOT, adapted, sized, rate, sigma
    real(kind=single), dimension(:,:), pointer :: aDR32, inIMG_RP, inMASK_RP, &
         inQPB_RP, inPROT_RP, inSWCX_RP, inEXP_RP, outADAPT_RP, outRATE_RP, outSIZE_RP, &
         outSIGMA_RP, inBKG_RP, inOOT_RP
    real(kind=double), dimension(:,:), pointer :: aDR64
    !
    withoot = .FALSE. ! for now
    oot     = .FALSE. ! for now
    !
    ! Get the input parameters
    !
    withcomb = .FALSE.
    prefix   = ""
    prefix   = stringParameter(prefixpar)    
    if (index(prefix,"comb") .ne. 0) then
       withcomb = .TRUE.
       write(errstr,*)"Using input from combimage"
       call message(AppMsg,Always) ! change to Noisy???!!!
    else
       if (prefix .eq. "") then ! this should never happen since comb is default
          call error("prefixORcomb",&
               "You must enter either inst+expid OR comb for combimage inputs")
       end if
    endif
    !
    elow        = intParameter(elowpar)         ! IEL
    ehigh       = intParameter(ehighpar)        ! IEH
    write(celow,'(I5)') elow   ! character form of integer
    write(cehigh,'(I5)') ehigh ! character form of integer
    withpart    = booleanParameter(partpar)     ! IPART
    withprot    = booleanParameter(sppar)       ! IPROT
    withswcx    = booleanParameter(swcxpar)     ! ISWCX
    withmask    = .FALSE.
    withmask    = booleanParameter(maskpar)
    inMASK_F    = "" 
    if (withmask) inMASK_F   = stringParameter(maskfilepar) ! OLDNEW: MASKFILE
    withbin     = .FALSE.
    withbin     = booleanParameter(withbinpar)
    if (withbin) then
       binfact = intParameter(binfactpar)     ! OLDNEW: ibin
    else
       binfact = 1
    endif
    !
    !  SS: Enter number of counts for smoothing.  The program smoothes using 
    !  a circular region surrounding the pixel whose radius is increased
    !  until the selected number of counts are enclosed.
    !  
    withsmoo    = .FALSE.
    withsmoo    = booleanParameter(withsmoopar)
    if (withsmoo) smoofact = realParameter(smoocntspar)  ! OLDNEW: SN
    !
    !  SS: Enter the fractional limit for template, the program will exclude
    !  all pixels in the field where the mask value is less than the 
    !  fractional limit times the average non-zero value of the mask
    !
    maskthresh = realParameter(tmpar)        ! OLDNEW: temsc
    !    withplots  = .FALSE.
    !    withplots  = booleanParameter(plotspar)  ! do QDP plots?
    if ((.not. withbin) .and. (.not. withsmoo)) then
       call error("noBinNoSmooth",&
            "withbinning=F and withsmoothing=F is not allowed. Program would do nothing.")
    endif
    !
    ! create the filenames, fill the arrays
    ! call check_clobber ???!!!
    ! OUTPUT FILES:
    ! OLDNEW: adapt-350-1100.fits/mos1S001-adaptimsky-350-1100.fits
    ! OLDNEW: size-350-1100.fits/mos1S001-sizeimsky-350-1100.fits
    ! BHP: or just do all of them by default???!!!
    suffix = ".fits"
    style = "adaptimsky" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,outADAPT_F) ! output ADAPT
    suffix = ".fits"
    style = "sizeimsky" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,outSIZE_F) ! output SIZE
    suffix = ".fits"
    style = "rateimsky" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,outRATE_F) ! output RATE
    suffix = ".fits"
    style = "sigimsky" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,outSIGMA_F) ! output SIGMA
    !
    ! Set the output QDP plot filenames
    !
    suffix = ".qdp"
    style = "radfilt" !OLDNEW: mos1S001-radial-filt-350-1100.qdp/mos1S001-radfilt-350-1100.qdp
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,outRADQ_F) ! output rad QDP
    style = "size" !OLDNEW: mos1S001-size-350-1100.adp/mos1S001-sizeimsky-350-1100.adp
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,outSIZEQ_F) ! output sz QDP
    !
    ! INPUT FILES:
    ! Get the FOV image and keywords
    ! OLDNEW: comb-obj-im-350-1100.fits/comb-fovim-350-1100.fits use sky???!!! or...
    ! OLDNEW: mos1S001-obj-im-350-1100.fits/mos1S001-fovimsky-350-1100.fits
    suffix = ".fits"
    style = "fovimsky" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inIMG_F) ! input IMAGE
    write(errstr,*)'***** Reading file: ',trim(adjustl(inIMG_F))
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_open_set(inIMG_F,inIMG_S,1,status,warndie) ! goflib/gofio_mod
    type = 3 ! image
    ! extract image array and coordinate keys (these are in gofio/goflib also)
    !
    ! BHP: ordinary type FOVIM is Int32, otherwise all other IMGs used in adapt are
    !      Real32
    ! fill the counts array:
    allocate(counts(xSize,ySize))
    counts = 0.
    if (withcomb) then ! comb type files are R*4, non-comb files are I*4
       call read_2Darray(inIMG_S,inIMG_A,inIMG_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             counts(i,j) = inIMG_RP(i-1,j-1)
          enddo
       enddo
    else
       call read_2Darray(inIMG_S,inIMG_A,aDR32,aDR64,aDI16,inIMG_JP,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             counts(i,j) = real(inIMG_JP(i-1,j-1))
          enddo
       enddo
    endif
    call coord_keys (inIMG_S, ra_pntO, dec_pntO, pa_pntO, crvalO, cdeltO, &
         cunitO, ctypeO, crpixO, crotaO, radecsysO, warndie)
    !
    ! BHP: get the PN OOT, apply to counts array (if applicable)
    ! comb-oot???!!!
    ! OLDNEW: pnS003-obj-im-oot-350-1100.fits/pnS003-fovimootsky-350-1100.fits
    if (trim(adjustl(inst)) .eq. 'PN') then ! always???!!! or param withoot???!!!
       write(errstr,*)'***** Reading file: ',trim(adjustl(pnOOT_F))
       call message(AppMsg,Always)
       oot = .TRUE.
       suffix = ".fits"
       style = "fovimootsky" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,pnOOT_F) ! input IMAGE
       write(errstr,*)'PN OOT file = ',trim(adjustl(pnOOT_F))
       call message(AppMsg,Always)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(pnOOT_F,pnOOT_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,pnOOT_S,pnOOT_T,pnOOT_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(pnOOT_S,pnOOT_A,aDR32,aDR64,aDI16,pnOOT_JP,xSizeD,ySizeD,typeA)
       ! SS: This is a bit painful, but we need to know the SUBMODE
       !     keyword for the proper scaling of the OOT events
       ! BHP: Not anymore. Just need the smode from the header, then call
       !      getootsca from goflib...
       call getootsca(smode,ootsca,warndie) ! goflib/gofutils OOT scale factor
       ! apply oot to counts array (no need to put into own array, never gets used again!)
       do i=1,xSize
          do j=1,ySize
             counts(i,ii) = counts(i,ii) - ootsca * real(pnOOT_JP(i-1,ii-1))
          enddo
       enddo
       call release(pnOOT_S)
       withoot = .TRUE.
    endif
    ! Get the exposure map and keywords
    ! OLDNEW: comb-exp-im-350-1100.fits/comb-fovim-350-1100.fits sky???!!!
    ! OLDNEW: mos1S001-exp-im-350-1100.fits/mos1S001-expimsky-350-1100.fits
    style = "expimsky" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inEXP_F) ! input exposure map
    allocate(expose(xSize,ySize))
    expose = 0.
    write(errstr,*)'***** Reading file: ',trim(adjustl(inEXP_F))
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_open_set(inEXP_F,inEXP_S,1,status,warndie) ! goflib/gofio_mod
    type = 3 ! image
    ! extract image array and coordinate keys (these are in gofio/goflib also)
    call read_header_info(type,inEXP_S,inEXP_T,inEXP_A,inst,obsid,expid,&
         smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
    call read_2Darray(inEXP_S,inEXP_A,inEXP_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
    call coord_keys (inIMG_S, ra_pntE, dec_pntE, pa_pntE, crvalE, cdeltE, &
         cunitE, ctypeE, crpixE, crotaE, radecsysE, warndie)
    do i=1,xSize
       do j=1,ySize
          expose(i,j) = inEXP_RP(i-1,j-1)
       enddo
    enddo
    call release(inEXP_S)
    ! Optional images...
    ! Get the mask and keywords (if selected)
    ! OLDNEW: comb-mask-im-350-1100.fits/mos1S001-maskimsky-350-1100.fits
    ! OLDNEW: mos1S001-mask-im-350-1100.fits/mos1S001-maskimsky-350-1100.fits
    allocate(mask(xSize,ySize))
    mask = 0. ! mask is usually an Int32, but why set it to EXPOSE if no mask???!!!
    if (withmask) then
       if (index(inMASK_F,"default") .ne. 0) then
          style = "maskimsky" 
          call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inMASK_F) ! input mask
       endif
       write(errstr,*)'***** Reading file: ',trim(adjustl(inMASK_F))
       call message(AppMsg,Noisy)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(inMASK_F,inMASK_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,inMASK_S,inMASK_T,inMASK_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(inMASK_S,inMASK_A,aDR32,aDR64,aDI16,inMASK_JP,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             if (typeA .eq. 2) then ! add other possibilities or unnecessary???!!!
                mask(i,j) = real(inMASK_JP(i-1,j-1))
             else
                mask(i,j) = adR32(i-1,j-1)
             endif
          enddo
       enddo
       call release(inMASK_S)
    else
       write(errstr,*)'***** No MASK entered, setting mask = exposure array'
       call message(AppMsg,Always)
       do i=1,xSize
          do j=1,ySize
             mask(i,j) = expose(i,j) ! why???!!!
          enddo
       enddo
    endif
    ! Get particle background backim image and keywords (if selected)
    ! OLDNEW: comb-back-im-350-1100.fits/comb-bkgim-350-1100.fits sky???!!!
    ! OLDNEW: mos1S001-back-im-sky-350-1100.fits/mos1S001-bkgimsky-350-1100.fits
    allocate(back(xSize,ySize))
    back = 0.
    if (withpart) then 
       style = "bkgimsky" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inBKG_F) ! input background map
       write(errstr,*)'***** Reading file: ',trim(adjustl(inBKG_F))
       call message(AppMsg,Always)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(inBKG_F,inBKG_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,inBKG_S,inBKG_T,inBKG_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(inBKG_S,inBKG_A,inBKG_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             back(i,j) = inBKG_RP(i-1,j-1)
             ! verified exact: if (back(i,j).gt.0) write(23,*)i,j,back(i,j)
          enddo
       enddo
       call release(inBKG_S)
    endif
    ! Get proton-produced soft proton image and keywords
    ! OLDNEW: comb-prot-im-350-1100.fits/comb-protim-350-1100.fits sky???!!!    
    ! OLDNEW: mos1S001-prot-im-sky-350-1100.fits/mos1S001-protimsky-350-1100.fits
    allocate(prot(xSize,ySize))
    prot = 0.
    if (withprot) then 
       style = "protimsky" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inPROT_F) ! input SP map
       write(errstr,*)'***** Reading file: ',trim(adjustl(inPROT_F))
       call message(AppMsg,Always)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(inPROT_F,inPROT_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,inPROT_S,inPROT_T,inPROT_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(inPROT_S,inPROT_A,inPROT_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             prot(i,j) = inPROT_RP(i-1,j-1)
          enddo
       enddo
       call release(inPROT_S)
    endif
    ! Get swcx-produced image and keywords
    ! OLDNEW: comb-swcx-im-350-1100.fits/comb-swcxim-350-1100.fits sky???!!!    
    ! OLDNEW: mos1S001-swcx-im-sky-350-1100.fits/mos1S001-swcximsky-350-1100.fits
    allocate(swcx(xSize,ySize))
    swcx = 0.
    if (withswcx) then 
       style = "swcximsky" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inSWCX_F) ! input SWCX map
       write(errstr,*)'***** Reading file: ',trim(adjustl(inSWCX_F))
       call message(AppMsg,Always)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(inSWCX_F,inSWCX_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,inSWCX_S,inSWCX_T,inSWCX_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(inSWCX_S,inSWCX_A,inSWCX_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             swcx(i,j) = inSWCX_RP(i-1,j-1)
          enddo
       enddo
       call release(inSWCX_S)
    endif
    call printParams(inIMG_F,inEXP_F,inBKG_F,outADAPT_F,outRATE_F,outSIZE_F,outSIGMA_F,&
         outSIZEQ_F,outRADQ_F,withcomb,celow,cehigh,withpart,withprot,withswcx,withmask,&
         inMASK_F,withbin,binfact,withsmoo,smoofact,maskthresh)
    !
    !  SS: Find out how many counts we have to play with and the exposure center
    ! 
    imcounts   = 0.0 ! RNC
    qpbcounts  = 0.0 ! RNB
    protcounts = 0.0 ! RNP
    swcxcounts = 0.0 ! RNS
    expcounts  = 0.0 ! REX
    ex = 0.0
    do i=1,xSize
       do ii=1,ySize
          imcounts   = imcounts + counts(i,ii)
          qpbcounts  = qpbcounts + back(i,ii)
          protcounts = protcounts + prot(i,ii)
          swcxcounts = swcxcounts + swcx(i,ii)
          expcounts  = expcounts + expose(i,ii)
          back(i,ii) = back(i,ii) + prot(i,ii) + swcx(i,ii)
          if(expose(i,ii) .gt. ex) then
             ex = expose(i,ii)
             xcen = i
             ycen = ii
          endif
       enddo
    enddo
    ! BHP: these are exactly the same, old v new
    write(errstr,*) "Total Counts   : ", imcounts 
    call message(AppMsg,Always)
    write(errstr,*) "Exposure       : ", expcounts
    call message(AppMsg,Always)
    write(errstr,*) "QPB Counts     : ", qpbcounts
    if (withpart) call message(AppMsg,Always)
    write(errstr,*) "Proton Counts  : ", protcounts
    if (withprot) call message(AppMsg,Always)
    write(errstr,*) "SWCX Counts    : ", swcxcounts
    if (withswcx) call message(AppMsg,Always)
    !
    ioff = mod(xSize,binfact)/2 ! 900
    xdim = xSize/binfact
    ydim = ySize/binfact ! OLDNEW: ibin
    sca = -1.0 / (cdeltO(1)*cdeltO(2)) ! or cdeltE???!!!
    !
    !
    ! SS: Read in the template
    ! SS: Use the exposure maps as the template
    ! OLD: IF(MASKFILE .EQ. '      ') THEN
    !
    ! SS: Return the mask (BHP: already done above)
    ! SS: Set the template limit
    !
    j = 0
    tlim = 0.
    do i=1,xSize
       do ii=1,ySize
          if(mask(i,ii) .gt. 0) then
             tlim = tlim + mask(i,ii)
             j = j + 1
          endif
       enddo
    enddo
    tlim = maskthresh * tlim / j ! OLDNEW: TEMSC*TLIM/J
    write(errstr,*) "tlim = ",tlim,xcen,ycen
    call message(AppMsg,Always)
    !
    !  Make a radial profile 
    !
    allocate (cts(50),bck(50),expos(50))
    allocate (x(50),y(50),y2(50),ys(50),n(50))
    cts   = 0.
    bck   = 0.
    expos = 0.
    x     = 0.
    y     = 0.
    y2    = 0.
    ys    = 0.
    n     = 0
    ir    = 0
    r     = 0.
    do i=1,xSize
       do ii=1,ySize
          if(mask(i,ii) .gt. 0.0) then
             r = 2.5*sqrt((i-xcen)**2 + (ii-ycen)**2)/60.0
             ir = int(r) + 1
             if(ir .le. 50) then
                cts(ir)  = cts(ir) + counts(i,ii)
                bck(ir)  = bck(ir) + back(i,ii)
                expos(ir) = expos(ir) + expose(i,ii) ! OLDNEW: expo/expos
                n(ir)    = n(ir) + 1
             endif
          endif
       enddo
    enddo
    xmax = 0.0
    ymax = 0.0
    do i=1,30
       if(expos(i) .gt. 0.0) then
          x(i)  = i - 0.5
          y(i)  = sca * (cts(i) - bck(i))/expos(i)
          y2(i) = sca * cts(i)/expos(i)
          ys(i) = sca * sqrt(cts(i))/expos(i)
          if(cts(i) .gt. 100.0) then
             xmax = i
             if(y2(i) .gt. ymax) ymax = y2(i) + ys(i)
          endif
          write(errstr,2990) x(i), y(i), ys(i), &
               n(i), cts(i)/n(i), &
               bck(i)/n(i), expos(i)/n(i)
2990      format(1x,3f12.5,i12,2f12.5,f12.2)
          call message(appmsg,Verbose)
       else
          x(i) = 0.
          y(i) = 0.
          y2(i) = 0.
          ys(i) = 0.
       endif
    enddo
    ymax = 1.05 * ymax
    !
    ! create and write output QDP file (RADIAL)
    ! e.g. mos1S001-radfilt-350-1100.qdp exactly the same old v new
    !
    !    if (withplots) then ! removed for SAS-21, can return???!!!
    warndie = .FALSE.
    call check_clobber(outRADQ_F,warndie)
    outlun = GetLun()
    open(unit=outlun,status='new',form='formatted',&
         file=trim(adjustl(outRADQ_F)),iostat=ios)
    if (ios .ne. 0) then
       call warning("badQDP",&
            "QDP output file could not be opened for write, no RADIAL QDP output done")
    else
       write(outlun,*) 'VIEWPORT 0.15 0.15 0.9 0.85'
       write(outlun,*) 'READ SERR 2'
       write(outlun,*) 'FONT ROMAN'
       write(outlun,*) 'CSIZE 1.8'
       write(outlun,*) 'LA F'
       write(outlun,*) 'LA X Radius (arc minutes)'
       write(outlun,*) 'LA Y Surface Brightness (counts/s/deg\u2\d)'
       write(outlun,*) 'COL 1 ON 2'
       write(outlun,*) 'COL 2 ON 3'
       write(outlun,*) 'R X 0 ', xmax
       write(outlun,*) 'R Y 0 ', ymax
       write(outlun,*) 'LINE STEP'
       do i=1,50 ! why 50 arrays only GO to 30???!!!
          if(ys(i) .gt. 0.0) then
             write(outlun,*) x(i), y(i), ys(i), y2(i)
          endif
       enddo
       close(outlun)
    endif
    !    end if
    !
    !  SS: Set the binning limits
    ! BHP: binfact defaults to 1 if no binning, but does binning anyway (see below)
    !if (withbin) then
    i2 = xSize / binfact ! OLDNEW: 900/ibin
    allocate(rate(i2,i2),sigma(i2,i2))
    rate = 0.
    sigma = 0.
    !
    write(errstr,*) "Start the binning"
    call message(AppMsg,Always)
    b = 0.
    c = 0.
    e = 0.
    ! from here ----
    do i=1,i2
       j2 = i*binfact + ioff
       j1 = j2 - binfact + 1
       do ii=1,i2
          jj2 = ii*binfact + ioff
          jj1 = jj2 - binfact + 1
          b = 0.
          c = 0.
          e = 0.
          do j=j1,j2
             do jj=jj1,jj2
                if(mask(j,jj) .gt. tlim) then
                   b = b + back(j,jj)
                   c = c + counts(j,jj)
                   e = e + expose(j,jj)
                endif
             enddo
          enddo
          ! to here ----exactly the same in old adapt and binimage
          if (withsmoo) then ! not done in binimage
             back(i,ii)   = b
             counts(i,ii) = c
             expose(i,ii) = e
          endif
          ! BHP: none of the below was ever done in adapt
          if(e .gt. 0.0) then
             r = sca*(c - b)/e
             if(c .gt. 0.0) then
                s = sca*sqrt(c)/e
             else
                s = 0.0
             endif
          else
             r = 0.0
             s = 0.0
          endif
          rate(i,ii) = r
          sigma(i,ii) = s
       enddo
    enddo
    !endif
    !
    !  Now smooth the data (can you smooth rate/sigma images???!!!)
    !  BHP: bin_image did NOT allow smoothing
    !
    if (withsmoo) then ! BHP: don't bother if not true
       b = 0.
       c = 0.
       e = 0.
       write(errstr,*) "Start the smoothing, SmoothFactor = ",smoofact,' xSize/ySize = ',xSize,ySize
       call message(AppMsg,Always)
       !
       allocate(adapted(i2,i2),sized(i2,i2),xh(1000),yh(1000)) ! why 1000???!!!
       adapted = 0.
       sized = 0.
       xh = 0.
       yh = 0.
       do i=1,1000
          xh(i) = -0.1 + i/15.0 ! hardcode this???!!!
          yh(i) = 0.
       end do
       do i=1,i2
          if(mod(i,25) .eq. 0) then
             write(errstr,*) "Step: ", i, "out of ",i2
             call message(AppMsg,Noisy)
          endif
          do ii=1,i2
             if(expose(i,ii) .gt. 0.) then
                b   = 0.
                c   = 0.
                e   = 0.
                num = 0
                k   = -1
                rad = 0.0
                do while (c .lt. smoofact)
                   blast = b
                   clast = c
                   elast = e
                   numlast = num
                   k = k + 1
                   if(k .eq. 1) then
                      rad = 1.5*1.5
                   elseif(k .eq. 2) then
                      rad = 2.5*2.5
                   elseif(k .eq. 3) then
                      rad = 3.2*3.2
                   else
                      rad = k*k*1.05*1.05
                   endif
                   j1 = i - k
                   if(j1 .lt. 1) j1 = 1
                   j2 = i + k
                   if(j2 .gt. i2) j2 = i2
                   jj1 = ii - k
                   if(jj1 .lt. 1) jj1 = 1
                   jj2 = ii + k
                   if(jj2 .gt. i2) jj2 = i2
                   b = 0.
                   c = 0.
                   e = 0.
                   num = 0
                   do j=j1,j2
                      do jj=jj1,jj2
                         trad = (j-i)**2 + (jj-ii)**2
                         if((expose(j,jj) .gt. 0.) .and. &
                              (trad .le. rad)) then
                            if(rad .eq. 0.0) then
                               b = b + back(j,jj)
                               c = c + counts(j,jj)
                               e = e + expose(j,jj)
                            else
                               t = exp(-((trad/rad)**2))
                               b = b + back(j,jj) * t
                               c = c + counts(j,jj) * t
                               e = e + expose(j,jj) * t
                            endif
                            num = num + 1
                         endif
                      enddo
                   enddo
                enddo
                if((k .gt. 1) .and. (numlast .gt. 0.)) then
                   temp = sca * (clast - blast) / elast
                   adapted(i,ii) = temp
                elseif(num .gt. 0.) then
                   temp = sca * (c - b) / e
                   adapted(i,ii) = temp
                else
                   adapted(i,ii) = temp
                endif
                rad = 2.5*sqrt(rad)/60.0             
                sized(i,ii) = rad
                ir = 15*rad + 1
                yh(ir) = yh(ir) + 1
             endif
          enddo
       enddo
       ymax = 0.
       do i=1,100
          if(yh(i) .gt. ymax) ymax = yh(i)
       enddo
       !
       ! create and write output QDP files (SIZE)
       ! ONLY produce if withbin+withsmoo, which = withsmoothing
       if (withsmoo) then 
          warndie = .FALSE.
          call check_clobber(outSIZEQ_F,warndie)
          outlun = GetLun()
          open(unit=outlun,status='new',form='formatted',&
               file=trim(adjustl(outSIZEQ_F)),iostat=ios)
          if (ios .ne. 0) then
             call warning("badQDP",&
                  "QDP output file could not be opened for write, no SIZE QDP output done")
          else
             write(outlun,*) 'VIEWPORT 0.15 0.15 0.95 0.85'
             write(outlun,*) 'FONT ROMAN'
             write(outlun,*) 'CSIZE 1.8'
             write(outlun,*) 'LA T Smoothing FWHM Histogram'
             write(outlun,*) 'LA F'
             write(outlun,*) 'LA Y Number'
             write(outlun,*) 'LA X Conical Sampling FWHM (arcmin)'
             write(outlun,*) 'R X 0 ', 5.
             write(outlun,*) 'R Y 0 ', 1.05*ymax
             write(outlun,*) 'LINE STEP'
             write(outlun,*) 'COL 1 ON 2'
             do i=1,50
                write(outlun,*) xh(i), yh(i)
             enddo
             close(outlun)
          endif
       endif
    endif
    !
    ! Open the new file size.fits and adapt.fits, write out
    !
    cdelt(1) = binfact * cdeltE(1) ! exactly the same adapt/binimage
    cdelt(2) = binfact * cdeltE(2) ! for all 4 images 
    crpix(1) = xdim / 2 + 0.5
    crpix(2) = ydim / 2 + 0.5
    !
    status = 0
    warndie = .TRUE.
    type = 3 ! 3 = image
    name = "PRIMARY"
    if (withsmoo) then 
       call check_clobber(outADAPT_F,warndie) ! croak if SAS_CLOBBER is set
       ! 3 = CREATE
       call check_open_set(outADAPT_F,outADAPT_S,3,status,warndie) ! goflib/gofio_mod
       outADAPT_A = addArray(outADAPT_S, trim(name), REAL32, (/xdim,ydim/))
       outADAPT_RP => real32Array2Data(outADAPT_A)    
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_clobber(outSIZE_F,warndie)
       ! 3 = CREATE
       call check_open_set(outSIZE_F,outSIZE_S,3,status,warndie) ! goflib/gofio_mod
       outSIZE_A = addArray(outSIZE_S, trim(name), REAL32, (/xdim,ydim/))
       outSIZE_RP => real32Array2Data(outSIZE_A)
    else
       call check_clobber(outRATE_F,warndie) ! croak if SAS_CLOBBER is set
       ! 3 = CREATE
       call check_open_set(outRATE_F,outRATE_S,3,status,warndie) ! goflib/gofio_mod
       outRATE_A = addArray(outRATE_S, trim(name), REAL32, (/xdim,ydim/))
       outRATE_RP => real32Array2Data(outRATE_A)    
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_clobber(outSIGMA_F,warndie)
       ! 3 = CREATE
       call check_open_set(outSIGMA_F,outSIGMA_S,3,status,warndie) ! goflib/gofio_mod
       outSIGMA_A = addArray(outSIGMA_S, trim(name), REAL32, (/xdim,ydim/))
       outSIGMA_RP => real32Array2Data(outSIGMA_A)
    endif
    !
    ! add comments from originals ???!!!
    !
    if (withsmoo) then
       call updateKeywords(outADAPT_A,radecsysE,crpix,ctypeE,crvalE,cdelt)
       call updateKeywords(outSIZE_A,radecsysE,crpix,ctypeE,crvalE,cdelt)
    else
       call updateKeywords(outRATE_A,radecsysE,crpix,ctypeE,crvalE,cdelt)
       call updateKeywords(outSIGMA_A,radecsysE,crpix,ctypeE,crvalE,cdelt)       
    endif

    ! fill the final array pointers opened above
    do i=1,xdim
       do j=1,ydim
          if (withsmoo) then 
             outADAPT_RP(i-1,j-1) = adapted(i,j)
             outSIZE_RP(i-1,j-1)  = sized(i,j)
          else
             outRATE_RP(i-1,j-1)  = rate(i,j)
             outSIGMA_RP(i-1,j-1) = sigma(i,j)
          endif
       enddo
    enddo
    ! close and seal the output files
    if (withsmoo) then 
       call release(outADAPT_S)
       call release(outSIZE_S)
    else
       call release(outRATE_S)
       call release(outSIGMA_S)
    endif
    
    return

  end subroutine binadapt
  !
  ! Subroutine section
  !
  subroutine makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,file)
    !
    character(len=*), intent(in) :: prefix, suffix, style, celow, cehigh
    character(len=256), intent(out) :: file
    character(len=3) :: pre, sky
    !
    logical :: withcomb, oot
    !
    !
    if (withcomb) then 
       file = "comb-" // trim(adjustl(style)) //'-' // trim(adjustl(celow)) &
            // '-' // trim(adjustl(cehigh)) // trim(adjustl(suffix))
    else       
       pre = 'mos'
       if ((prefix(1:1) .ne. '1') .and. (prefix(1:1) .ne. '2')) then
          pre = 'pn'
       endif
       if (oot) then 
          file = trim(adjustl(pre)) // trim(adjustl(prefix)) // &
               '-' // trim(adjustl(style)) // '-' // &
               trim(adjustl(celow)) // '-' // trim(adjustl(cehigh)) &
               // trim(adjustl(suffix))
       else
          file = trim(adjustl(pre)) // trim(adjustl(prefix)) // &
               '-' // trim(adjustl(style)) // '-' // &
               trim(adjustl(celow)) // '-' // trim(adjustl(cehigh)) &
               // trim(adjustl(suffix))
       endif
    endif

    return

  end subroutine makeFileNames
  !
  !
  !
  subroutine printParams(inIMG_F,inEXP_F,inBKG_F,outADAPT_F,outRATE_F,outSIZE_F,outSIGMA_F,&
       outSIZEQ_F,outRADQ_F,withcomb,celow,cehigh,withpart,withprot,withswcx,withmask,&
       inMASK_F,withbin,binfact,withsmoo,smoofact,maskthresh)
    !
    character(len=*) :: inIMG_F, inEXP_F, inBKG_F, outADAPT_F, outRATE_F, outSIZE_F,&
         outSIGMA_F, outSIZEQ_F, outRADQ_F, inMASK_F, celow, cehigh
    character(len=4) :: cbin
    !
    integer :: binfact
    !
    logical, intent(in) :: withcomb, withpart, withprot, withswcx, withmask, &
         withbin, withsmoo
    !
    real :: smoofact, maskthresh
    write(cbin,'(I4)') binfact   ! character form

    write(errstr,*) "Input parameters:"
    call message(AppMsg,Always)
    write(errstr,*) "Input IMAGE filename             : ",trim(inIMG_F)    ! OLDNEW: INFILE
    call message(AppMsg,Always)
    if(withcomb) then
       write(errstr,*) "Use combimage-produced files?    : ",withcomb 
       call message(AppMsg,Always)
    endif
    write(errstr,*) "Input EXP IMAGE filename         : ",trim(inEXP_F)    ! OLDNEW: INFILE!
    call message(AppMsg,Always)
    if (withsmoo) then 
       write(errstr,*) "Output IMAGE (adapted) filename  : ",trim(outADAPT_F) ! OLDNEW: INFILE!
       call message(AppMsg,Always)
       write(errstr,*) "Output SIZE FITS file            : ",trim(outSIZE_F)  ! OLDNEW: INFILES!
       call message(AppMsg,Always)
       write(errstr,*) "Output SIZE QDP file             : ",trim(outSIZEQ_F) ! OLDNEW: INFILES!
       call message(AppMsg,Always)
       write(errstr,*) "Output RADIAL QDP file           : ",trim(outRADQ_F)  ! OLDNEW: INFILES!
       call message(AppMsg,Always)
    endif
    if (withbin .and. (.not.withsmoo)) then
       write(errstr,*) "Output RATE IMAGE filename       : ",trim(outRATE_F)   ! OLDNEW: INFILE!
       call message(AppMsg,Always)
       write(errstr,*) "Output SIGMA IMAGE filename      : ",trim(outSIGMA_F)   ! OLDNEW: INFILE!
       call message(AppMsg,Always)
       write(errstr,*) "Output RADIAL QDP file           : ",trim(outRADQ_F)  ! OLDNEW: INFILES!
       call message(AppMsg,Always)
    endif
    write(errstr,*) "Apply source removal mask?       : ",withmask
    call message(AppMsg,Always)
    if (withmask) then
       write(errstr,*) "Source removal mask              : ",trim(adjustl(inMASK_F))
       call message(AppMsg,Always)
    endif
    write(errstr,*) "Selected energy lower limit      : ",trim(adjustr(celow))
    call message(AppMsg,Always)
    write(errstr,*) "Selected energy upper limit      : ",trim(adjustr(cehigh))
    call message(AppMsg,Always)
    write(errstr,*) "Apply particle (QPB) count data? : ",withpart
    call message(AppMsg,Always)
    write(errstr,*) "Input BKG IMAGE filename         : ",trim(inBKG_F)    ! OLDNEW: INFILE!
    if (withpart) call message(AppMsg,Always)
    write(errstr,*) "Apply soft proton count data?    : ",withprot
    if (withprot) call message(AppMsg,Always)
    write(errstr,*) "Apply SWCX count data?           : ",withswcx
    if (withswcx) call message(AppMsg,Always)
    write(errstr,*) "Bin the data?                    : ",withbin
    call message(AppMsg,Always)
    write(errstr,*) "Bin factor                       :  ",trim(adjustl(cbin))
    if (withbin)  call message(AppMsg,Always)
    write(errstr,*) "Smooth the data?                 : ",withsmoo
    call message(AppMsg,Always)
    write(errstr,*) "Number of counts for smoothing   : ", smoofact
    if (withsmoo) call message(AppMsg,Always)       
    write(errstr,*) "Fractional limit for template    : ", maskthresh
    call message(AppMsg,Always)
    !write(errstr,*) "Make QDP diagnostic plots?       : ",withplots
    !call message(AppMsg,Always)

    return

  end subroutine printparams
  !
  !
  !
  subroutine updateKeywords(out_A,radecsys,crpix,ctype,crval,cdelt)

    type(ArrayT)   :: out_A
    character(len=10) :: radecsys, ctype(2)
    real(kind=double) :: crval(2), crpix(2), cdelt(2)

    call setAttribute(out_A, 'CONTENT', "IMAGE")
    call setAttribute(out_A, 'TELESCOP', "XMM", "mission name")
    call setAttribute(out_A, 'INSTRUME', "EPIC","instrument name")
    call setAttribute(out_A, 'OBS_MODE', "POINTING","obs mode: POINTING, SLEW, OR SCAN")
    call setAttribute(out_A, 'RADECSYS', radecsys,"Equitorial system reference")
    call setAttribute(out_A, 'EQUINOX', 2000, "Equinox")
    !call setAttribute(out_A, 'BZERO', 0., "") how to write???!!!
    !call setAttribute(out_A, 'BSCALE', 1., "")
    !call setAttribute(out_A, 'BUNIT', "c/s/deg2", "Units of data")
    call setAttribute(out_A, 'CRPIX1', crpix(1), "Reference pixel")
    call setAttribute(out_A, 'CTYPE1', ctype(1), "Projection")
    call setAttribute(out_A, 'CRVAL1', crval(1), "Central Longitude")
    call setAttribute(out_A, 'CDELT1', cdelt(1), "WCS pixel size")
    call setAttribute(out_A, 'CUNIT1', "deg", "Units of coordinate")
    call setAttribute(out_A, 'CRPIX2', crpix(2), "Reference pixel")
    call setAttribute(out_A, 'CTYPE2', ctype(2), "Projection")
    call setAttribute(out_A, 'CRVAL2', crval(2), "Central Latitude")    
    call setAttribute(out_A, 'CDELT2', cdelt(2), "WCS pixel size")
    call setAttribute(out_A, 'CUNIT2', "deg", "Units of coordinate")

    return

  end subroutine updateKeywords

end module binadapt_mod
